
# Code from Chapter 2 of Machine Learning: An Algorithmic Perspective
# by Stephen Marsland (http://seat.massey.ac.nz/personal/s.r.marsland/MLBook.html)

# You are free to use, change, or redistribute the code in any way you wish for
# non-commercial purposes, but please maintain the name of the original author.
# This code comes with no warranty of any kind.

# Stephen Marsland, 2008

import numpy as np

class pcn:
    """ A basic Perceptron"""
    
    def __init__(self,inputs,targets,activationFunction='l'):
        """ Constructor """
        # Set up network size
        if np.ndim(inputs)>1:
            self.nIn = np.shape(inputs)[1]
        else: 
            self.nIn = 1
    
        if np.ndim(targets)>1:
            self.nOut = np.shape(targets)[1]
        else:
            self.nOut = 1

        self.inputs = inputs
        self.targets = targets

        self.nData = np.shape(inputs)[0]
    
        # Initialise network
        self.weights = np.random.rand(self.nIn+1,self.nOut)*0.1-0.05

        # Activation function and its derivative
        functions = {'l': lambda x: np.where(x>0,1,0),
                     's': lambda x: 1./(1.+np.exp(-x))}
        functionsDeriv = {'l': lambda x: 1.,
                          's': lambda x: x*(1.-x)}
        if activationFunction in functions:
            self.f = functions[activationFunction]
            self.fd = functionsDeriv[activationFunction]
        else:
            self.f = functions['l']
            self.fd = functionsDeriv['l']

    def pcntrain(self,eta,nIterations,verbose=False):
        """ Train the thing """    
        # Add the inputs that match the bias node
        inputs = np.concatenate((self.inputs,-np.ones((self.nData,1))),axis=1)
        targets = self.targets
        # Training
        change = range(self.nData)

        for n in range(nIterations):
            
            self.outputs = self._pcnfwd(inputs);
            self.weights += eta*np.dot(np.transpose(inputs),self.fd(self.outputs)*(targets-self.outputs))
            if verbose:
                print "Iteration: ", n
		print self.weights
			
		activations = self._pcnfwd(inputs)
		print "Final outputs are:"
		print activations
        
            # Randomise order of inputs
            np.random.shuffle(change)
            inputs = inputs[change,:]
            targets = targets[change,:]

    def _pcnfwd(self,inputs):
        """ Run the network forward """

        outputs = np.dot(inputs,self.weights)

        # Threshold the outputs
        return self.f(outputs)

    def output(self, inputs):
        """ Run the trained network """

        inputs = np.concatenate((inputs,-np.ones((np.shape(inputs)[0],1))),axis=1)
        outputs = self._pcnfwd(inputs)

        # Threshold the outputs
        return self.f(outputs)


    def confmat(self,inputs,targets):
        """Confusion matrix"""

        nClasses = np.shape(targets)[1]

        if nClasses==1:
            nClasses = 2
            outputs = self.output(inputs)
        else:
            # Add the inputs that match the bias node
            inputs = np.concatenate((inputs,-np.ones((self.nData,1))),axis=1)
        
            outputs = np.dot(inputs,self.weights)
        
            # 1-of-N encoding
            outputs = np.argmax(outputs,1)
            targets = np.argmax(targets,1)

        outputs = np.where(outputs>0.6,1,0)

        cm = np.zeros((nClasses,nClasses))
        for i in range(nClasses):
            for j in range(nClasses):
                cm[i,j] = np.sum(np.where(outputs==i,1,0)*np.where(targets==j,1,0))

        print cm
        print np.trace(cm)/np.sum(cm)
        
def logic():
    """ Run AND, OR and XOR logic functions"""
    a = np.array([[0,0,0],[0,1,0],[1,0,0],[1,1,1]],dtype=float)
    b = np.array([[0,0,0],[0,1,1],[1,0,1],[1,1,1]],dtype=float)
    c = np.array([[0,0,0],[0,1,1],[1,0,1],[1,1,0]],dtype=float)
    
    q = pcn(a[:,0:2],a[:,2:],'s')
    q.pcntrain(0.25,30)
    q.confmat(a[:,0:2],a[:,2:])
    
    q = pcn(b[:,0:2],b[:,2:])
    q.pcntrain(0.25,20)
    q.confmat(b[:,0:2],b[:,2:])

    p = pcn(c[:,0:2],c[:,2:],'s')
    p.pcntrain(0.05,100)
    p.confmat(c[:,0:2],c[:,2:])
    
if __name__ == "__main__":
    logic()
